// Copyright (C) 2013 Yusuke Suzuki <utatane.tea@gmail.com>
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//   * Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
//   * Redistributions in binary form must reproduce the above copyright
//     notice, this list of conditions and the following disclaimer in the
//     documentation and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"use strict";

const expect = require("chai").expect;
const espree = require("./util/espree");
const analyze = require("..").analyze;

describe("optimistic", () => {
    it("direct call to eval", () => {
        const ast = espree(`
            function outer() {
                eval(str);
                var i = 20;
                function inner() {
                    i;
                }
            }
        `);

        const scopes = analyze(ast, { optimistic: true }).scopes;

        expect(scopes.map(scope => scope.variables.map(variable => variable.name))).to.be.eql(
            [
                [
                    "outer"
                ],
                [
                    "arguments",
                    "i",
                    "inner"
                ],
                [
                    "arguments"
                ]
            ]
        );
    });

    it("with statement", () => {
        const ast = espree(`
            function outer() {
                eval(str);
                var i = 20;
                with (obj) {
                    i;
                }
            }
        `, "script");

        const scopes = analyze(ast, { optimistic: true }).scopes;

        expect(scopes.map(scope => scope.variables.map(variable => variable.name))).to.be.eql(
            [
                [
                    "outer"
                ],
                [
                    "arguments",
                    "i"
                ],
                [
                ]
            ]
        );
    });
});
